'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-17 21:24:18
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 18:50:16
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import time
import torch
from torch import Tensor
from torch.utils.checkpoint import checkpoint
import numpy as np
import matplotlib.pyplot as plt
from typing import Optional,List

##########################################################################
#                          FD Coefficient     
##########################################################################
@torch.jit.script
def DiffCoef(order: int, grid_scheme: str) -> torch.Tensor:
    """
    Calculate the differential coefficients for a given order and grid scheme.

    Parameters:
        order (int): The order of the coefficients.
        grid_scheme (str): The grid scheme, either 'r' for rational or 's' for symmetric.

    Returns:
        torch.Tensor: The calculated differential coefficients.
    """
    # Initialize the coefficient array
    C = torch.zeros(order)

    # Determine the coefficient matrix and right-hand side based on the grid scheme
    if grid_scheme == 'r':
        B = torch.cat((torch.tensor([1 / 2]), torch.zeros(order - 1)))
        A = torch.empty((order, order), dtype=torch.float32)

        for i in range(order):
            for j in range(order):
                A[i, j] = (j + 1) ** (2 * (i + 1) - 1)

    elif grid_scheme == 's':
        B = torch.cat((torch.tensor([1.0]), torch.zeros(order - 1)))
        A = torch.empty((order, order), dtype=torch.float32)

        for i in range(order):
            for j in range(order):
                A[i, j] = (2 * (j + 1) - 1) ** (2 * (i + 1) - 1)

    else:
        raise ValueError("Invalid grid scheme. Use 'r' for rational or 's' for symmetric.")

    # Calculate the differential coefficients using PyTorch
    C = torch.linalg.inv(A) @ B

    return C



##########################################################################
#                          FD Operator     
##########################################################################

@torch.jit.script
def Dxfm_4(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return fdc[0] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj]) + \
           fdc[1] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 1])
            
@torch.jit.script
def Dzfm_4(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return fdc[0] * (a[:, ii + 1, :][:, :, jj] - a[:, ii, :][:, :, jj]) + \
           fdc[1] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 1, :][:, :, jj])

@torch.jit.script
def Dxbm_4(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii, :][:, :, jj - 1]) + \
           fdc[1] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj - 2])

@torch.jit.script
def Dzbm_4(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
           fdc[1] * (a[:, ii + 1, :][:, :, jj] - a[:, ii - 2, :][:, :, jj])

@torch.jit.script
def Dxfm_6(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return fdc[0] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj]) + \
           fdc[1] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 1]) + \
           fdc[2] * (a[:, ii, :][:, :, jj + 3] - a[:, ii, :][:, :, jj - 2])

@torch.jit.script
def Dzfm_6(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii + 1, :][:, :, jj] - a[:, ii, :][:, :, jj]) + \
            fdc[1] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
            fdc[2] * (a[:, ii + 3, :][:, :, jj] - a[:, ii - 2, :][:, :, jj])

@torch.jit.script
def Dxbm_6(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii, :][:, :, jj - 1]) + \
            fdc[1] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj - 2]) + \
            fdc[2] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 3])

@torch.jit.script
def Dzbm_6(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
            fdc[1] * (a[:, ii + 1, :][:, :, jj] - a[:, ii - 2, :][:, :, jj]) + \
            fdc[2] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 3, :][:, :, jj])

@torch.jit.script
def Dxfm_8(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj]) + \
            fdc[1] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 1]) + \
            fdc[2] * (a[:, ii, :][:, :, jj + 3] - a[:, ii, :][:, :, jj - 2]) + \
            fdc[3] * (a[:, ii, :][:, :, jj + 4] - a[:, ii, :][:, :, jj - 3])

@torch.jit.script
def Dzfm_8(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii + 1, :][:, :, jj] - a[:, ii, :][:, :, jj]) + \
            fdc[1] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
            fdc[2] * (a[:, ii + 3, :][:, :, jj] - a[:, ii - 2, :][:, :, jj]) + \
            fdc[3] * (a[:, ii + 4, :][:, :, jj] - a[:, ii - 3, :][:, :, jj])

@torch.jit.script
def Dxbm_8(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii, :][:, :, jj - 1]) + \
            fdc[1] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj - 2]) + \
            fdc[2] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 3]) + \
            fdc[3] * (a[:, ii, :][:, :, jj + 3] - a[:, ii, :][:, :, jj - 4])

@torch.jit.script
def Dzbm_8(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
            fdc[1] * (a[:, ii + 1, :][:, :, jj] - a[:, ii - 2, :][:, :, jj]) + \
            fdc[2] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 3, :][:, :, jj]) + \
            fdc[3] * (a[:, ii + 3, :][:, :, jj] - a[:, ii - 4, :][:, :, jj])

@torch.jit.script
def Dxfm_10(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj]) + \
            fdc[1] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 1]) + \
            fdc[2] * (a[:, ii, :][:, :, jj + 3] - a[:, ii, :][:, :, jj - 2]) + \
            fdc[3] * (a[:, ii, :][:, :, jj + 4] - a[:, ii, :][:, :, jj - 3]) + \
            fdc[4] * (a[:, ii, :][:, :, jj + 5] - a[:, ii, :][:, :, jj - 4])

@torch.jit.script
def Dzfm_10(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii + 1, :][:, :, jj] - a[:, ii, :][:, :, jj]) + \
            fdc[1] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
            fdc[2] * (a[:, ii + 3, :][:, :, jj] - a[:, ii - 2, :][:, :, jj]) + \
            fdc[3] * (a[:, ii + 4, :][:, :, jj] - a[:, ii - 3, :][:, :, jj]) + \
            fdc[4] * (a[:, ii + 5, :][:, :, jj] - a[:, ii - 4, :][:, :, jj])

@torch.jit.script
def Dxbm_10(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii, :][:, :, jj - 1]) + \
            fdc[1] * (a[:, ii, :][:, :, jj + 1] - a[:, ii, :][:, :, jj - 2]) + \
            fdc[2] * (a[:, ii, :][:, :, jj + 2] - a[:, ii, :][:, :, jj - 3]) + \
            fdc[3] * (a[:, ii, :][:, :, jj + 3] - a[:, ii, :][:, :, jj - 4]) + \
            fdc[4] * (a[:, ii, :][:, :, jj + 4] - a[:, ii, :][:, :, jj - 5])

@torch.jit.script
def Dzbm_10(a: torch.Tensor, fdc: torch.Tensor, ii: torch.Tensor, jj: torch.Tensor) -> torch.Tensor:
    return  fdc[0] * (a[:, ii, :][:, :, jj] - a[:, ii - 1, :][:, :, jj]) + \
            fdc[1] * (a[:, ii + 1, :][:, :, jj] - a[:, ii - 2, :][:, :, jj]) + \
            fdc[2] * (a[:, ii + 2, :][:, :, jj] - a[:, ii - 3, :][:, :, jj]) + \
            fdc[3] * (a[:, ii + 3, :][:, :, jj] - a[:, ii - 4, :][:, :, jj]) + \
            fdc[4] * (a[:, ii + 4, :][:, :, jj] - a[:, ii - 5, :][:, :, jj])


##########################################################################
#                          FD padding Coefficient     
##########################################################################
@torch.jit.script
def pad_torchSingle(data: torch.Tensor, pml: int, fs_offset: int, free_surface: bool = True, device: torch.device = torch.device("cpu")) -> torch.Tensor:
    """
    Description:
        Pad a 2D tensor with specified padding layers (PML) and optional free surface condition.

    Parameters:
        - data (torch.Tensor): Input tensor to be padded, with shape (nz, nx).
        - pml (int): Number of padding layers to apply (Perfectly Matched Layer).
        - fs_offset (int): Offset for the free surface condition.
        - free_surface (bool): Flag indicating whether to apply free surface conditions. Default is True.
        - device (torch.device): Device to allocate the padded tensor on. Default is CPU.

    Returns:
        - torch.Tensor: Padded tensor with shape (nz_pml, nx_pml).
    """
    nz, nx = data.shape
    
    # Calculate new dimensions
    nx_pml = nx + 2 * pml
    nz_pml = nz + (pml + fs_offset if free_surface else 2 * pml + fs_offset)
    
    # Initialize padded tensor
    cc = torch.zeros((nz_pml, nx_pml), device=device)
    
    if free_surface:
        # Copy data to padded tensor with free surface condition
        cc[fs_offset:fs_offset + nz, pml:pml + nx] = data
        cc[:fs_offset, pml:pml + nx] = data[0, :].unsqueeze(0).expand(fs_offset, -1)  # Top padding
    else:
        # Copy data to padded tensor without free surface
        cc[fs_offset + pml:fs_offset + pml + nz, pml:pml + nx] = data
        cc[:fs_offset + pml, pml:pml + nx] = data[0, :].unsqueeze(0).expand(fs_offset + pml, -1)  # Top padding

    # Bottom padding
    cc[nz_pml - pml:, pml:pml + nx] = data[-1, :].unsqueeze(0).expand(pml, -1)
    
    # Left and right padding
    cc[:, :pml] = cc[:, pml:pml + 1].expand(-1, pml)  # Left padding
    cc[:, nx_pml - pml:] = cc[:, nx_pml - pml - 1:nx_pml - pml].expand(-1, pml)  # Right padding

    return cc

##########################################################################
#                        step forward Modeling    
##########################################################################
@torch.jit.script
def step_forward_PML_4order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                bcx:Tensor,bcz:Tensor,                                                      # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx_x:Tensor,txx_z:Tensor,tzz_x:Tensor,tzz_z:Tensor,txz_x:Tensor,txz_z:Tensor,txx:Tensor,tzz:Tensor,txz:Tensor, # intermedia variable
                vx_x:Tensor,vx_z:Tensor,vz_x:Tensor,vz_z:Tensor,vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    """Description:
        Simulates the time-stepping of the wavefield in a 2D elastic medium using finite difference methods with Perfectly Matched Layer (PML) boundary conditions.
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Grid arrangement%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %                                                                                           %
            %                        txx,tzz__________ vx ___________ txx,tzz------>                    %
            %                lbd,mu  |                        bh                            |           %
            %                             |                         |                             |     %
            %                             |                         |                             |     %
            %                             |                         |                             |     %
            %                        vz  |____________txz                           |                   %
            %                       bv  |                       muvh                        |           %
            %                             |                                                        |    %
            %                             |                                                        |    %
            %                             |                                                        |    %
            %                 txx,tzz  |____________________________|                                   %
            %                          |                                                                %
            %                         \|/                                                               %
            %                                                                                           %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Grid arrangement%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    Prameters:
    --------------
        - M (int): The finite difference order. Typically 4, 6, 8, or 10, determining the accuracy of spatial derivatives.
        - free_surface (bool): Indicates whether to impose a free surface condition at the top boundary.
        - nx (int), nz (int): Number of grid points in the x and z directions, respectively, defining the simulation domain.
        - dx (float), dz (float): Spatial step sizes in the x and z directions.
        - nabc (int): The number of grid points used for the PML absorbing boundary.
        - src_x (Tensor), src_z (Tensor), src_n (int): Source positions in x and z directions, and the number of sources.
        - dt (float): Time step size.
        - src_v (Tensor): Source time function values.
        - MT (Tensor): Moment tensor for the source mechanism.
        - rcv_x (Tensor), rcv_z (Tensor), rcv_n (int): Receiver positions in x and z directions, and the number of receivers.
        - bcx (Tensor), bcz (Tensor): Damping profiles for the absorbing boundary conditions in x and z directions.
        - lam (Tensor), lamu (Tensor): Lame parameters (for P-wave and S-wave velocities).
        - C11, C13, C15, C33, C35, C55 (Tensor): Elastic moduli for the anisotropic medium.
        - bx (Tensor), bz (Tensor): Density-related coefficients in x and z directions.
        - txx_x, txx_z, tzz_x, tzz_z, txz_x, txz_z, txx, tzz, txz (Tensor): Stress tensors and their derivatives in x and z directions.
        - vx_x, vx_z, vz_x, vz_z, vx, vz (Tensor): Velocity components and their derivatives in x and z directions.
        - device (torch.device): The computing device (CPU or GPU).
        - dtype (torch.dtype): The data type (e.g., float32) used for the computations.

    Returns:
    --------------
        A tuple containing the updated stress and velocity tensors as well as recorded receiver waveforms and forward wavefields.
    """
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset

    # PML dimensions with free surface condition
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz           = vx.clone()    , vz.clone()    , txx.clone(), tzz.clone(), txz.clone()
    vx_x, vz_x, txx_x, tzz_x, txz_x = vx_x.clone()  , vz_x.clone()  , txx_x.clone(), tzz_x.clone(), txz_x.clone()
    vx_z, vz_z, txx_z, tzz_z, txz_z = vx_z.clone()  , vz_z.clone()  , txx_z.clone(), tzz_z.clone(), txz_z.clone()

    # Initialize recorded waveforms
    rcv_txx, rcv_tzz, rcv_txz, rcv_vx, rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    forward_wavefield_txx, forward_wavefield_tzz, forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx, forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    idx_j_plus1  = slice(j_start+1,j_end+1)
    idx_j_minus1 = slice(j_start-1,j_end-1)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Damping factors for PML
    pmlxd = 1 + 0.5 * dt * bcx[idx_i, idx_j]
    pmlxn = 1 - 0.5 * dt * bcx[idx_i, idx_j]
    pmlzd = 1 + 0.5 * dt * bcz[idx_i, idx_j]
    pmlzn = 1 - 0.5 * dt * bcz[idx_i, idx_j]
    
    # Precompute constants for the forward loop
    dt_dx = dt / dx
    dt_dz = dt / dz
    pmlx_inv = 1.0 / pmlxd
    pmlz_inv = 1.0 / pmlzd
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    half_MT = MT[src_idx] / 3  # Half of the moment tensor
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_4
    Dzfm = Dzfm_4
    Dxbm = Dxbm_4
    Dzbm = Dzbm_4
    
    # moment tensor source implementation
    for t in range(nt):
        
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)

        # Update stress fields
        txx_x[:, idx_i, idx_j] = (pmlxn * txx_x[:, idx_i, idx_j] + dt_dx * (C11[idx_i, idx_j] * dxbm_vx + C15[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        txx_z[:, idx_i, idx_j] = (pmlzn * txx_z[:, idx_i, idx_j] + dt_dz * (C15[idx_i, idx_j] * dzbm_vx + C13[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        tzz_x[:, idx_i, idx_j] = (pmlxn * tzz_x[:, idx_i, idx_j] + dt_dx * (C13[idx_i, idx_j] * dxbm_vx + C35[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        tzz_z[:, idx_i, idx_j] = (pmlzn * tzz_z[:, idx_i, idx_j] + dt_dz * (C35[idx_i, idx_j] * dzbm_vx + C33[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        txz_x[:, idx_i, idx_j] = (pmlxn * txz_x[:, idx_i, idx_j] + dt_dx * (C15[idx_i, idx_j] * dxfm_vx + C55[idx_i, idx_j] * dxfm_vz)) * pmlx_inv
        txz_z[:, idx_i, idx_j] = (pmlzn * txz_z[:, idx_i, idx_j] + dt_dz * (C55[idx_i, idx_j] * dzfm_vx + C35[idx_i, idx_j] * dzfm_vz)) * pmlz_inv

        # Add moment tensor source
        if len(src_v.shape) == 1:
            txx_x[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            txx_z[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            txz_x[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
            txz_z[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
        else:
            txx_x[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            txx_z[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            txz_x[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]
            txz_z[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]

        # Combine x and z components of stress
        txx[:] = txx_x + txx_z
        tzz[:] = tzz_x + tzz_z
        txz[:] = txz_x + txz_z

        # Apply free surface boundary conditions
        if free_surface:
            tzz[:, h-1, :] = 0
            tzz[:, h-2, :] = -tzz[:, h, :]
            txz[:, h-2, :] = -txz[:, h-1, :]
            txz[:, h-3, :] = -txz[:, h, :]

        # Compute velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx_x[:, idx_i, idx_j] = (pmlxn * vx_x[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dxfm_txx / dx) / pmlxd
        vx_z[:, idx_i, idx_j] = (pmlzn * vx_z[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dzbm_txz / dz) / pmlzd
        vz_x[:, idx_i, idx_j] = (pmlxn * vz_x[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dxbm_txz / dx) / pmlxd
        vz_z[:, idx_i, idx_j] = (pmlzn * vz_z[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dzfm_tzz / dz) / pmlzd
        vx[:] = vx_x + vx_z
        vz[:] = vz_x + vz_z

        # Apply free surface boundary conditions for velocity
        # if free_surface:
        #     vz[:, h-2, idx_j] = vz[:, h-1, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-1, idx_j] - vx[:, h-1, idx_j_minus1])
        #     vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
        #     vz[:, h-3, idx_j] = vz[:, h-2, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-2, idx_j] - vx[:, h-2, idx_j_minus1])
        
        if free_surface:
            vz[:, h-2, idx_j] = vz[:, h-1, idx_j]
            vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
            vz[:, h-3, idx_j] = vz[:, h-2, idx_j]

        # -----------------------------------------------------------
        #                   Receiver Observation
        # -----------------------------------------------------------
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]

        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx  += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz  += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()

    return txx_x,txx_z,tzz_x,tzz_z,txz_x,txz_z,txx,tzz,txz,\
            vx_x,vx_z,vz_x,vz_z,vx,vz,\
            rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz

@torch.jit.script
def step_forward_PML_6order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                bcx:Tensor,bcz:Tensor,                                                      # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx_x:Tensor,txx_z:Tensor,tzz_x:Tensor,tzz_z:Tensor,txz_x:Tensor,txz_z:Tensor,txx:Tensor,tzz:Tensor,txz:Tensor, # intermedia variable
                vx_x:Tensor,vx_z:Tensor,vz_x:Tensor,vz_z:Tensor,vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset

    # PML dimensions with free surface condition
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz           = vx.clone()    , vz.clone()    , txx.clone(), tzz.clone(), txz.clone()
    vx_x, vz_x, txx_x, tzz_x, txz_x = vx_x.clone()  , vz_x.clone()  , txx_x.clone(), tzz_x.clone(), txz_x.clone()
    vx_z, vz_z, txx_z, tzz_z, txz_z = vx_z.clone()  , vz_z.clone()  , txx_z.clone(), tzz_z.clone(), txz_z.clone()

    # Initialize recorded waveforms
    rcv_txx, rcv_tzz, rcv_txz, rcv_vx, rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    forward_wavefield_txx, forward_wavefield_tzz, forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx, forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    idx_j_plus1 = slice(j_start+1,j_end+1)
    idx_j_minus1 = slice(j_start-1,j_end-1)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Damping factors for PML
    pmlxd = 1 + 0.5 * dt * bcx[idx_i, idx_j]
    pmlxn = 1 - 0.5 * dt * bcx[idx_i, idx_j]
    pmlzd = 1 + 0.5 * dt * bcz[idx_i, idx_j]
    pmlzn = 1 - 0.5 * dt * bcz[idx_i, idx_j]
    
    # Precompute constants for the forward loop
    dt_dx = dt / dx
    dt_dz = dt / dz
    pmlx_inv = 1.0 / pmlxd
    pmlz_inv = 1.0 / pmlzd
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    half_MT = MT[src_idx] / 2  # Half of the moment tensor
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_6
    Dzfm = Dzfm_6
    Dxbm = Dxbm_6
    Dzbm = Dzbm_6
    
    # moment tensor source implementation
    for t in range(nt):
        
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)

        # Update stress fields
        txx_x[:, idx_i, idx_j] = (pmlxn * txx_x[:, idx_i, idx_j] + dt_dx * (C11[idx_i, idx_j] * dxbm_vx + C15[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        txx_z[:, idx_i, idx_j] = (pmlzn * txx_z[:, idx_i, idx_j] + dt_dz * (C15[idx_i, idx_j] * dzbm_vx + C13[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        tzz_x[:, idx_i, idx_j] = (pmlxn * tzz_x[:, idx_i, idx_j] + dt_dx * (C13[idx_i, idx_j] * dxbm_vx + C35[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        tzz_z[:, idx_i, idx_j] = (pmlzn * tzz_z[:, idx_i, idx_j] + dt_dz * (C35[idx_i, idx_j] * dzbm_vx + C33[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        txz_x[:, idx_i, idx_j] = (pmlxn * txz_x[:, idx_i, idx_j] + dt_dx * (C15[idx_i, idx_j] * dxfm_vx + C55[idx_i, idx_j] * dxfm_vz)) * pmlx_inv
        txz_z[:, idx_i, idx_j] = (pmlzn * txz_z[:, idx_i, idx_j] + dt_dz * (C55[idx_i, idx_j] * dzfm_vx + C35[idx_i, idx_j] * dzfm_vz)) * pmlz_inv

        # Add moment tensor source
        if len(src_v.shape) == 1:
            txx_x[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            txx_z[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            txz_x[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
            txz_z[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
        else:
            txx_x[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            txx_z[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            txz_x[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]
            txz_z[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]

        # Combine x and z components of stress
        txx[:] = txx_x + txx_z
        tzz[:] = tzz_x + tzz_z
        txz[:] = txz_x + txz_z

        # Apply free surface boundary conditions
        if free_surface:
            tzz[:, h-1, :] = 0
            tzz[:, h-2, :] = -tzz[:, h, :]
            txz[:, h-2, :] = -txz[:, h-1, :]
            txz[:, h-3, :] = -txz[:, h, :]

        # Compute velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx_x[:, idx_i, idx_j] = (pmlxn * vx_x[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dxfm_txx / dx) / pmlxd
        vx_z[:, idx_i, idx_j] = (pmlzn * vx_z[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dzbm_txz / dz) / pmlzd
        vz_x[:, idx_i, idx_j] = (pmlxn * vz_x[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dxbm_txz / dx) / pmlxd
        vz_z[:, idx_i, idx_j] = (pmlzn * vz_z[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dzfm_tzz / dz) / pmlzd
        vx[:] = vx_x + vx_z
        vz[:] = vz_x + vz_z

        # Apply free surface boundary conditions for velocity
        # if free_surface:
        #     vz[:, h-2, idx_j] = vz[:, h-1, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-1, idx_j] - vx[:, h-1, idx_j_minus1])
        #     vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
        #     vz[:, h-3, idx_j] = vz[:, h-2, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-2, idx_j] - vx[:, h-2, idx_j_minus1])
        
        if free_surface:
            vz[:, h-2, idx_j] = vz[:, h-1, idx_j]
            vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
            vz[:, h-3, idx_j] = vz[:, h-2, idx_j]

        # -----------------------------------------------------------
        #                   Receiver Observation
        # -----------------------------------------------------------
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]

        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx  += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz  += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        
    return txx_x,txx_z,tzz_x,tzz_z,txz_x,txz_z,txx,tzz,txz,\
            vx_x,vx_z,vz_x,vz_z,vx,vz,\
            rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz

@torch.jit.script
def step_forward_PML_8order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                bcx:Tensor,bcz:Tensor,                                                      # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx_x:Tensor,txx_z:Tensor,tzz_x:Tensor,tzz_z:Tensor,txz_x:Tensor,txz_z:Tensor,txx:Tensor,tzz:Tensor,txz:Tensor, # intermedia variable
                vx_x:Tensor,vx_z:Tensor,vz_x:Tensor,vz_z:Tensor,vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset

    # PML dimensions with free surface condition
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz           = vx.clone()    , vz.clone()    , txx.clone(), tzz.clone(), txz.clone()
    vx_x, vz_x, txx_x, tzz_x, txz_x = vx_x.clone()  , vz_x.clone()  , txx_x.clone(), tzz_x.clone(), txz_x.clone()
    vx_z, vz_z, txx_z, tzz_z, txz_z = vx_z.clone()  , vz_z.clone()  , txx_z.clone(), tzz_z.clone(), txz_z.clone()

    # Initialize recorded waveforms
    rcv_txx, rcv_tzz, rcv_txz, rcv_vx, rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    forward_wavefield_txx, forward_wavefield_tzz, forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx, forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    idx_j_plus1 = slice(j_start+1,j_end+1)
    idx_j_minus1 = slice(j_start-1,j_end-1)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Damping factors for PML
    pmlxd = 1 + 0.5 * dt * bcx[idx_i, idx_j]
    pmlxn = 1 - 0.5 * dt * bcx[idx_i, idx_j]
    pmlzd = 1 + 0.5 * dt * bcz[idx_i, idx_j]
    pmlzn = 1 - 0.5 * dt * bcz[idx_i, idx_j]
    
    # Precompute constants for the forward loop
    dt_dx = dt / dx
    dt_dz = dt / dz
    pmlx_inv = 1.0 / pmlxd
    pmlz_inv = 1.0 / pmlzd
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    half_MT = MT[src_idx] / 2  # Half of the moment tensor
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_8
    Dzfm = Dzfm_8
    Dxbm = Dxbm_8
    Dzbm = Dzbm_8
    
    # moment tensor source implementation
    for t in range(nt):
        
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)

        # Update stress fields
        txx_x[:, idx_i, idx_j] = (pmlxn * txx_x[:, idx_i, idx_j] + dt_dx * (C11[idx_i, idx_j] * dxbm_vx + C15[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        txx_z[:, idx_i, idx_j] = (pmlzn * txx_z[:, idx_i, idx_j] + dt_dz * (C15[idx_i, idx_j] * dzbm_vx + C13[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        tzz_x[:, idx_i, idx_j] = (pmlxn * tzz_x[:, idx_i, idx_j] + dt_dx * (C13[idx_i, idx_j] * dxbm_vx + C35[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        tzz_z[:, idx_i, idx_j] = (pmlzn * tzz_z[:, idx_i, idx_j] + dt_dz * (C35[idx_i, idx_j] * dzbm_vx + C33[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        txz_x[:, idx_i, idx_j] = (pmlxn * txz_x[:, idx_i, idx_j] + dt_dx * (C15[idx_i, idx_j] * dxfm_vx + C55[idx_i, idx_j] * dxfm_vz)) * pmlx_inv
        txz_z[:, idx_i, idx_j] = (pmlzn * txz_z[:, idx_i, idx_j] + dt_dz * (C55[idx_i, idx_j] * dzfm_vx + C35[idx_i, idx_j] * dzfm_vz)) * pmlz_inv

        # Add moment tensor source
        if len(src_v.shape) == 1:
            txx_x[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            txx_z[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            txz_x[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
            txz_z[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
        else:
            txx_x[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            txx_z[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            txz_x[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]
            txz_z[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]

        # Combine x and z components of stress
        txx[:] = txx_x + txx_z
        tzz[:] = tzz_x + tzz_z
        txz[:] = txz_x + txz_z

        # Apply free surface boundary conditions
        if free_surface:
            tzz[:, h-1, :] = 0
            tzz[:, h-2, :] = -tzz[:, h, :]
            txz[:, h-2, :] = -txz[:, h-1, :]
            txz[:, h-3, :] = -txz[:, h, :]

        # Compute velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx_x[:, idx_i, idx_j] = (pmlxn * vx_x[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dxfm_txx / dx) / pmlxd
        vx_z[:, idx_i, idx_j] = (pmlzn * vx_z[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dzbm_txz / dz) / pmlzd
        vz_x[:, idx_i, idx_j] = (pmlxn * vz_x[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dxbm_txz / dx) / pmlxd
        vz_z[:, idx_i, idx_j] = (pmlzn * vz_z[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dzfm_tzz / dz) / pmlzd
        vx[:] = vx_x + vx_z
        vz[:] = vz_x + vz_z

        # Apply free surface boundary conditions for velocity
        # if free_surface:
        #     vz[:, h-2, idx_j] = vz[:, h-1, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-1, idx_j] - vx[:, h-1, idx_j_minus1])
        #     vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
        #     vz[:, h-3, idx_j] = vz[:, h-2, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-2, idx_j] - vx[:, h-2, idx_j_minus1])
        
        if free_surface:
            vz[:, h-2, idx_j] = vz[:, h-1, idx_j]
            vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
            vz[:, h-3, idx_j] = vz[:, h-2, idx_j]

        # -----------------------------------------------------------
        #                   Receiver Observation
        # -----------------------------------------------------------
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]

        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx  += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz  += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        
    return txx_x,txx_z,tzz_x,tzz_z,txz_x,txz_z,txx,tzz,txz,\
            vx_x,vx_z,vz_x,vz_z,vx,vz,\
            rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz

@torch.jit.script
def step_forward_PML_10order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                bcx:Tensor,bcz:Tensor,                                                      # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx_x:Tensor,txx_z:Tensor,tzz_x:Tensor,tzz_z:Tensor,txz_x:Tensor,txz_z:Tensor,txx:Tensor,tzz:Tensor,txz:Tensor, # intermedia variable
                vx_x:Tensor,vx_z:Tensor,vz_x:Tensor,vz_z:Tensor,vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset

    # PML dimensions with free surface condition
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz           = vx.clone()    , vz.clone()    , txx.clone(), tzz.clone(), txz.clone()
    vx_x, vz_x, txx_x, tzz_x, txz_x = vx_x.clone()  , vz_x.clone()  , txx_x.clone(), tzz_x.clone(), txz_x.clone()
    vx_z, vz_z, txx_z, tzz_z, txz_z = vx_z.clone()  , vz_z.clone()  , txx_z.clone(), tzz_z.clone(), txz_z.clone()

    # Initialize recorded waveforms
    rcv_txx, rcv_tzz, rcv_txz, rcv_vx, rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device), torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    forward_wavefield_txx, forward_wavefield_tzz, forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx, forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device), torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    idx_j_plus1 = slice(j_start+1,j_end+1)
    idx_j_minus1 = slice(j_start-1,j_end-1)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Damping factors for PML
    pmlxd = 1 + 0.5 * dt * bcx[idx_i, idx_j]
    pmlxn = 1 - 0.5 * dt * bcx[idx_i, idx_j]
    pmlzd = 1 + 0.5 * dt * bcz[idx_i, idx_j]
    pmlzn = 1 - 0.5 * dt * bcz[idx_i, idx_j]
    
    # Precompute constants for the forward loop
    dt_dx = dt / dx
    dt_dz = dt / dz
    pmlx_inv = 1.0 / pmlxd
    pmlz_inv = 1.0 / pmlzd
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    half_MT = MT[src_idx] / 2  # Half of the moment tensor
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_10
    Dzfm = Dzfm_10
    Dxbm = Dxbm_10
    Dzbm = Dzbm_10
    
    # moment tensor source implementation
    for t in range(nt):
        
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)

        # Update stress fields
        txx_x[:, idx_i, idx_j] = (pmlxn * txx_x[:, idx_i, idx_j] + dt_dx * (C11[idx_i, idx_j] * dxbm_vx + C15[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        txx_z[:, idx_i, idx_j] = (pmlzn * txx_z[:, idx_i, idx_j] + dt_dz * (C15[idx_i, idx_j] * dzbm_vx + C13[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        tzz_x[:, idx_i, idx_j] = (pmlxn * tzz_x[:, idx_i, idx_j] + dt_dx * (C13[idx_i, idx_j] * dxbm_vx + C35[idx_i, idx_j] * dxbm_vz)) * pmlx_inv
        tzz_z[:, idx_i, idx_j] = (pmlzn * tzz_z[:, idx_i, idx_j] + dt_dz * (C35[idx_i, idx_j] * dzbm_vx + C33[idx_i, idx_j] * dzbm_vz)) * pmlz_inv
        txz_x[:, idx_i, idx_j] = (pmlxn * txz_x[:, idx_i, idx_j] + dt_dx * (C15[idx_i, idx_j] * dxfm_vx + C55[idx_i, idx_j] * dxfm_vz)) * pmlx_inv
        txz_z[:, idx_i, idx_j] = (pmlzn * txz_z[:, idx_i, idx_j] + dt_dz * (C55[idx_i, idx_j] * dzfm_vx + C35[idx_i, idx_j] * dzfm_vz)) * pmlz_inv

        # Add moment tensor source
        if len(src_v.shape) == 1:
            txx_x[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            txx_z[src_idx, src_z, src_x] += -half_MT[0, 0] * src_v[t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[2, 2] * src_v[t]
            txz_x[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
            txz_z[src_idx, src_z, src_x] += -half_MT[0, 2] * src_v[t]
        else:
            txx_x[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            txx_z[src_idx, src_z, src_x] += -half_MT[:, 0, 0] * src_v[src_idx, t]
            tzz_x[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            tzz_z[src_idx, src_z, src_x] += -half_MT[:, 2, 2] * src_v[src_idx, t]
            txz_x[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]
            txz_z[src_idx, src_z, src_x] += -half_MT[:, 0, 2] * src_v[src_idx, t]

        # Combine x and z components of stress
        txx[:] = txx_x + txx_z
        tzz[:] = tzz_x + tzz_z
        txz[:] = txz_x + txz_z

        # Apply free surface boundary conditions
        if free_surface:
            tzz[:, h-1, :] = 0
            tzz[:, h-2, :] = -tzz[:, h, :]
            txz[:, h-2, :] = -txz[:, h-1, :]
            txz[:, h-3, :] = -txz[:, h, :]

        # Compute velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx_x[:, idx_i, idx_j] = (pmlxn * vx_x[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dxfm_txx / dx) / pmlxd
        vx_z[:, idx_i, idx_j] = (pmlzn * vx_z[:, idx_i, idx_j] + dt * bx[idx_i, idx_j] * dzbm_txz / dz) / pmlzd
        vz_x[:, idx_i, idx_j] = (pmlxn * vz_x[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dxbm_txz / dx) / pmlxd
        vz_z[:, idx_i, idx_j] = (pmlzn * vz_z[:, idx_i, idx_j] + dt * bz[idx_i, idx_j] * dzfm_tzz / dz) / pmlzd
        vx[:] = vx_x + vx_z
        vz[:] = vz_x + vz_z

        # Apply free surface boundary conditions for velocity
        # if free_surface:
        #     vz[:, h-2, idx_j] = vz[:, h-1, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-1, idx_j] - vx[:, h-1, idx_j_minus1])
        #     vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
        #     vz[:, h-3, idx_j] = vz[:, h-2, idx_j]       + (lam[h-1, idx_j]/lamu[h-1, idx_j])*(vx[:, h-2, idx_j] - vx[:, h-2, idx_j_minus1])
        
        if free_surface:
            vz[:, h-2, idx_j] = vz[:, h-1, idx_j]
            vx[:, h-2, idx_j] = vz[:, h-2, idx_j_plus1] - vz[:, h-2, idx_j] + vz[:, h-1, idx_j_plus1] - vz[:, h-1, idx_j] + vx[:, h, idx_j]
            vz[:, h-3, idx_j] = vz[:, h-2, idx_j]

        # -----------------------------------------------------------
        #                   Receiver Observation
        # -----------------------------------------------------------
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]

        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx  += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz  += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        
    return txx_x,txx_z,tzz_x,tzz_z,txz_x,txz_z,txx,tzz,txz,\
            vx_x,vx_z,vz_x,vz_z,vx,vz,\
            rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz


##########################################################################
#                step forward Modeling :damping mode   
##########################################################################
@torch.jit.script
def step_forward_ABL_4order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                damp:Tensor,                                                                # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx:Tensor,tzz:Tensor,txz:Tensor,                                           # intermedia variable
                vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    """Description:
    --------------
        Perform a single time step in the simulation of wave propagation using a 4th order finite difference method.
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Grid arrangement%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %                                                                                           %
        %                        txx,tzz__________ vx ___________ txx,tzz------>                    %
        %                lbd,mu  |                        bh                            |           %
        %                             |                         |                             |     %
        %                             |                         |                             |     %
        %                             |                         |                             |     %
        %                        vz  |____________txz                           |                   %
        %                       bv  |                       muvh                        |           %
        %                             |                                                        |    %
        %                             |                                                        |    %
        %                             |                                                        |    %
        %                 txx,tzz  |____________________________|                                   %
        %                          |                                                                %
        %                         \|/                                                               %
        %                                                                                           %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Grid arrangement%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    Parameters:
    --------------
        - M (int): Order of the finite difference stencil.
        - free_surface (bool): Indicates whether free surface conditions should be applied.
        - nx (int): Number of grid points in the x-direction.
        - nz (int): Number of grid points in the z-direction.
        - dx (float): Grid spacing in the x-direction.
        - dz (float): Grid spacing in the z-direction.
        - nabc (int): Number of absorbing boundary condition layers.
        - src_x (Tensor): Source x-coordinates.
        - src_z (Tensor): Source z-coordinates.
        - src_n (int): Number of sources.
        - dt (float): Time step size.
        - src_v (Tensor): Source time function values.
        - MT (Tensor): Moment tensor representing the source.
        - rcv_x (Tensor): Receiver x-coordinates.
        - rcv_z (Tensor): Receiver z-coordinates.
        - rcv_n (int): Number of receivers.
        - damp (Tensor): Damping tensor for absorbing boundary condition.
        - lam (Tensor): First Lamé parameter.
        - lamu (Tensor): Second Lamé parameter.
        - C11, C13, C15, C33, C35, C55 (Tensor): Elastic moduli parameters.
        - bx (Tensor): Coefficient for x-direction velocity update.
        - bz (Tensor): Coefficient for z-direction velocity update.
        - txx (Tensor): Stress tensor component (xx).
        - tzz (Tensor): Stress tensor component (zz).
        - txz (Tensor): Stress tensor component (xz).
        - vx (Tensor): Velocity component in the x-direction.
        - vz (Tensor): Velocity component in the z-direction.
        - device (torch.device): Device to allocate tensors on (default is CPU).
        - dtype (torch.dtype): Data type for the tensors (default is float32).

    Returns:
    - Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]: 
      Updated stress tensors (txx, tzz, txz), updated velocity tensors (vx, vz),
      recorded waveforms (rcv_txx, rcv_tzz, rcv_txz, rcv_vx, rcv_vz),
      and forward wavefields (forward_wavefield_txx, forward_wavefield_tzz, forward_wavefield_txz, forward_wavefield_vx, forward_wavefield_vz).
    """
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset
    
    # Configure ABL dimensions based on boundary conditions
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz = vx.clone(), vz.clone(), txx.clone(), tzz.clone(), txz.clone()
    
    # Initialize recorded waveforms
    rcv_txx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_tzz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_txz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vx  = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vz  = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

    forward_wavefield_txx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_tzz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx  = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vz  = torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h       = NN + 1
    i_start = NN 
    i_end   = nz_pml - NN
    j_start = NN 
    j_end   = nx_pml - NN
    idx_i   = slice(i_start, i_end)
    idx_j   = slice(j_start, j_end)
    idx_j_plus1 = slice(j_start+1,j_end+1)
    idx_j_minus1 = slice(j_start-1,j_end-1)
    ii      = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj      = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_4
    Dzfm = Dzfm_4
    Dxbm = Dxbm_4
    Dzbm = Dzbm_4
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    scaling_factor = -1 / 2
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # moment tensor source implementation
    for t in range(nt):
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)
        
        txx[:, idx_i, idx_j] = txx[:, idx_i, idx_j] + dt*((C11[idx_i,idx_j]*dxbm_vx + C15[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C15[idx_i,idx_j]*dzbm_vx + C13[idx_i,idx_j]*dzbm_vz)/dz)
        tzz[:, idx_i, idx_j] = tzz[:, idx_i, idx_j] + dt*((C13[idx_i,idx_j]*dxbm_vx + C35[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C35[idx_i,idx_j]*dzbm_vx + C33[idx_i,idx_j]*dzbm_vz)/dz)
        txz[:, idx_i, idx_j] = txz[:, idx_i, idx_j] + dt*((C15[idx_i,idx_j]*dxfm_vx + C55[idx_i,idx_j]*dxfm_vz)/dx
                                                        + (C55[idx_i,idx_j]*dzfm_vx + C35[idx_i,idx_j]*dzfm_vz)/dz)
        
        # Add source
        if src_v.ndim == 1:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[0, 0] * src_v[t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[2, 2] * src_v[t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[0, 2] * src_v[t]
        else:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 0] * src_v[src_idx, t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 2, 2] * src_v[src_idx, t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 2] * src_v[src_idx, t]

        # topFs with the assumption of weak anisotropy near the surface
        if free_surface:
            tzz[:,h-1,:] = 0
            tzz[:,h-2,:] = -tzz[:,h,:]
            txz[:,h-2,:] = -txz[:,h-1,:]
            txz[:,h-3,:] = -txz[:,h,:]
        
        # Update velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx[:, idx_i, idx_j] += dt * bx[idx_i, idx_j] * (dxfm_txx / dx + dzbm_txz / dz)
        vz[:, idx_i, idx_j] += dt * bz[idx_i, idx_j] * (dxbm_txz / dx + dzfm_tzz / dz)

        # Apply free surface boundary conditions
        if free_surface:
            vz[:, h - 2, idx_j] = vz[:, h - 1, idx_j]
            vx[:, h - 2, idx_j] = vz[:, h - 2, idx_j_plus1] - vz[:, h - 2, idx_j] + vz[:, h - 1, idx_j_plus1] - vz[:, h - 1, idx_j] + vx[:, h, idx_j]
            vz[:, h - 3, idx_j] = vz[:, h - 2, idx_j]

        # Apply damping
        vx *= damp
        vz *= damp
        
        # Record receiver waveforms
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]
        
        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx  += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz  += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()

    return txx,tzz,txz,vx,vz,rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz

@torch.jit.script
def step_forward_ABL_6order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                damp:Tensor,                                                                # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx:Tensor,tzz:Tensor,txz:Tensor,                                           # intermedia variable
                vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset
    
    # Configure ABL dimensions based on boundary conditions
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz = vx.clone(), vz.clone(), txx.clone(), tzz.clone(), txz.clone()
    
    # Initialize recorded waveforms
    rcv_txx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_tzz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_txz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

    forward_wavefield_txx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_tzz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_6
    Dzfm = Dzfm_6
    Dxbm = Dxbm_6
    Dzbm = Dzbm_6
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    scaling_factor = -1 / 2
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # moment tensor source implementation
    for t in range(nt):
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)
        
        txx[:, idx_i, idx_j] = txx[:, idx_i, idx_j] + dt*((C11[idx_i,idx_j]*dxbm_vx + C15[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C15[idx_i,idx_j]*dzbm_vx + C13[idx_i,idx_j]*dzbm_vz)/dz)
        tzz[:, idx_i, idx_j] = tzz[:, idx_i, idx_j] + dt*((C13[idx_i,idx_j]*dxbm_vx + C35[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C35[idx_i,idx_j]*dzbm_vx + C33[idx_i,idx_j]*dzbm_vz)/dz)
        txz[:, idx_i, idx_j] = txz[:, idx_i, idx_j] + dt*((C15[idx_i,idx_j]*dxfm_vx + C55[idx_i,idx_j]*dxfm_vz)/dx
                                                        + (C55[idx_i,idx_j]*dzfm_vx + C35[idx_i,idx_j]*dzfm_vz)/dz)
        
        # Add source
        if src_v.ndim == 1:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[0, 0] * src_v[t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[2, 2] * src_v[t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[0, 2] * src_v[t]
        else:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 0] * src_v[src_idx, t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 2, 2] * src_v[src_idx, t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 2] * src_v[src_idx, t]

        # topFs with the assumption of weak anisotropy near the surface
        if free_surface:
            tzz[:,h-1,:] = 0
            tzz[:,h-2,:] = -tzz[:,h,:]
            txz[:,h-2,:] = -txz[:,h-1,:]
            txz[:,h-3,:] = -txz[:,h,:]
        
        # Update velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx[:, idx_i, idx_j] += dt * bx[idx_i, idx_j] * (dxfm_txx / dx + dzbm_txz / dz)
        vz[:, idx_i, idx_j] += dt * bz[idx_i, idx_j] * (dxbm_txz / dx + dzfm_tzz / dz)

        # Apply free surface boundary conditions
        if free_surface:
            vz[:, h - 2, idx_j] = vz[:, h - 1, idx_j]
            vx[:, h - 2, idx_j] = vz[:, h - 2, j_start + 1:j_end + 1] - vz[:, h - 2, idx_j] + vz[:, h - 1, j_start + 1:j_end + 1] - vz[:, h - 1, idx_j] + vx[:, h, idx_j]
            vz[:, h - 3, idx_j] = vz[:, h - 2, idx_j]

        # Apply damping
        vx *= damp
        vz *= damp
        
        # Record receiver waveforms
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]
        
        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx  += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz  += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()

    return txx,tzz,txz,vx,vz,rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz

@torch.jit.script
def step_forward_ABL_8order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                damp:Tensor,                                                                # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx:Tensor,tzz:Tensor,txz:Tensor,                                           # intermedia variable
                vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset
    
    # Configure ABL dimensions based on boundary conditions
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz = vx.clone(), vz.clone(), txx.clone(), tzz.clone(), txz.clone()
    
    # Initialize recorded waveforms
    rcv_txx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_tzz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_txz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

    forward_wavefield_txx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_tzz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_8
    Dzfm = Dzfm_8
    Dxbm = Dxbm_8
    Dzbm = Dzbm_8
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    scaling_factor = -1 / 2
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # moment tensor source implementation
    for t in range(nt):
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)
        
        txx[:, idx_i, idx_j] = txx[:, idx_i, idx_j] + dt*((C11[idx_i,idx_j]*dxbm_vx + C15[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C15[idx_i,idx_j]*dzbm_vx + C13[idx_i,idx_j]*dzbm_vz)/dz)
        tzz[:, idx_i, idx_j] = tzz[:, idx_i, idx_j] + dt*((C13[idx_i,idx_j]*dxbm_vx + C35[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C35[idx_i,idx_j]*dzbm_vx + C33[idx_i,idx_j]*dzbm_vz)/dz)
        txz[:, idx_i, idx_j] = txz[:, idx_i, idx_j] + dt*((C15[idx_i,idx_j]*dxfm_vx + C55[idx_i,idx_j]*dxfm_vz)/dx
                                                        + (C55[idx_i,idx_j]*dzfm_vx + C35[idx_i,idx_j]*dzfm_vz)/dz)
        
        # Add source
        if src_v.ndim == 1:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[0, 0] * src_v[t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[2, 2] * src_v[t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[0, 2] * src_v[t]
        else:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 0] * src_v[src_idx, t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 2, 2] * src_v[src_idx, t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 2] * src_v[src_idx, t]

        # topFs with the assumption of weak anisotropy near the surface
        if free_surface:
            tzz[:,h-1,:] = 0
            tzz[:,h-2,:] = -tzz[:,h,:]
            txz[:,h-2,:] = -txz[:,h-1,:]
            txz[:,h-3,:] = -txz[:,h,:]
        
        # Update velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx[:, idx_i, idx_j] += dt * bx[idx_i, idx_j] * (dxfm_txx / dx + dzbm_txz / dz)
        vz[:, idx_i, idx_j] += dt * bz[idx_i, idx_j] * (dxbm_txz / dx + dzfm_tzz / dz)

        # Apply free surface boundary conditions
        if free_surface:
            vz[:, h - 2, idx_j] = vz[:, h - 1, idx_j]
            vx[:, h - 2, idx_j] = vz[:, h - 2, j_start + 1:j_end + 1] - vz[:, h - 2, idx_j] + vz[:, h - 1, j_start + 1:j_end + 1] - vz[:, h - 1, idx_j] + vx[:, h, idx_j]
            vz[:, h - 3, idx_j] = vz[:, h - 2, idx_j]

        # Apply damping
        vx *= damp
        vz *= damp
        
        # Record receiver waveforms
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]
        
        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()

    return txx,tzz,txz,vx,vz,rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz

@torch.jit.script
def step_forward_ABL_10order(M:int,
                free_surface:bool,nx:int,nz:int,dx:float,dz:float,nabc:int,                 # basic settings
                src_x:Tensor,src_z:Tensor,src_n:int,dt:float,src_v:Tensor,MT:Tensor,        # source
                rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                        # receiver
                damp:Tensor,                                                                # absobing bondary condition
                lam:Tensor,lamu:Tensor,
                C11:Tensor,C13:Tensor,C15:Tensor,C33:Tensor,C35:Tensor,C55:Tensor,          # elastic moduli parameters
                bx:Tensor,bz:Tensor,                                                        
                txx:Tensor,tzz:Tensor,txz:Tensor,                                           # intermedia variable
                vx:Tensor,vz:Tensor,
                device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32
                ):
    nt = src_v.shape[-1]    # Number of time steps
    fs_offset = M // 2      # Finite difference stencil offset
    
    # Configure ABL dimensions based on boundary conditions
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # Clone tensors to avoid in-place modification
    vx, vz, txx, tzz, txz = vx.clone(), vz.clone(), txx.clone(), tzz.clone(), txz.clone()
    
    # Initialize recorded waveforms
    rcv_txx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_tzz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_txz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vx = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_vz = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

    forward_wavefield_txx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_tzz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_txz = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vx = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_vz = torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Finite difference order and indexing
    NN = M // 2
    fdc = DiffCoef(NN, 's')
    h = NN + 1
    i_start = NN 
    i_end = nz_pml - NN
    j_start = NN 
    j_end = nx_pml - NN
    idx_i = slice(i_start, i_end)
    idx_j = slice(j_start, j_end)
    ii = torch.arange(i_start, i_end, dtype=torch.long, device=device)  # z-axis range
    jj = torch.arange(j_start, j_end, dtype=torch.long, device=device)  # x-axis range
    
    # Finite difference operators for different axes
    Dxfm = Dxfm_10
    Dzfm = Dzfm_10
    Dxbm = Dxbm_10
    Dzbm = Dzbm_10
    
    # Source and receiver indices
    src_idx = list(range(src_n))
    rcv_idx = list(range(rcv_n))
    scaling_factor = -1 / 2
    
    # Adjust offset for free surface condition
    offset = fs_offset if free_surface else fs_offset + nabc
    
    # moment tensor source implementation
    for t in range(nt):
        # Compute stress components
        dxbm_vx = Dxbm(vx, fdc, ii, jj)
        dxbm_vz = Dxbm(vz, fdc, ii, jj)
        dzbm_vx = Dzbm(vx, fdc, ii, jj)
        dzbm_vz = Dzbm(vz, fdc, ii, jj)
        dxfm_vx = Dxfm(vx, fdc, ii, jj)
        dzfm_vx = Dzfm(vx, fdc, ii, jj)
        dxfm_vz = Dxfm(vz, fdc, ii, jj)
        dzfm_vz = Dzfm(vz, fdc, ii, jj)
        
        txx[:, idx_i, idx_j] = txx[:, idx_i, idx_j] + dt*((C11[idx_i,idx_j]*dxbm_vx + C15[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C15[idx_i,idx_j]*dzbm_vx + C13[idx_i,idx_j]*dzbm_vz)/dz)
        tzz[:, idx_i, idx_j] = tzz[:, idx_i, idx_j] + dt*((C13[idx_i,idx_j]*dxbm_vx + C35[idx_i,idx_j]*dxbm_vz)/dx
                                                        + (C35[idx_i,idx_j]*dzbm_vx + C33[idx_i,idx_j]*dzbm_vz)/dz)
        txz[:, idx_i, idx_j] = txz[:, idx_i, idx_j] + dt*((C15[idx_i,idx_j]*dxfm_vx + C55[idx_i,idx_j]*dxfm_vz)/dx
                                                        + (C55[idx_i,idx_j]*dzfm_vx + C35[idx_i,idx_j]*dzfm_vz)/dz)
        
        # Add source
        if src_v.ndim == 1:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[0, 0] * src_v[t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[2, 2] * src_v[t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[0, 2] * src_v[t]
        else:
            txx[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 0] * src_v[src_idx, t]
            tzz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 2, 2] * src_v[src_idx, t]
            txz[src_idx, src_z, src_x] += scaling_factor * MT[src_idx, 0, 2] * src_v[src_idx, t]

        # topFs with the assumption of weak anisotropy near the surface
        if free_surface:
            tzz[:,h-1,:] = 0
            tzz[:,h-2,:] = -tzz[:,h,:]
            txz[:,h-2,:] = -txz[:,h-1,:]
            txz[:,h-3,:] = -txz[:,h,:]
        
        # Update velocity components
        dxfm_txx = Dxfm(txx, fdc, ii, jj)
        dzbm_txz = Dzbm(txz, fdc, ii, jj)
        dxbm_txz = Dxbm(txz, fdc, ii, jj)
        dzfm_tzz = Dzfm(tzz, fdc, ii, jj)
        vx[:, idx_i, idx_j] += dt * bx[idx_i, idx_j] * (dxfm_txx / dx + dzbm_txz / dz)
        vz[:, idx_i, idx_j] += dt * bz[idx_i, idx_j] * (dxbm_txz / dx + dzfm_tzz / dz)

        # Apply free surface boundary conditions
        if free_surface:
            vz[:, h - 2, idx_j] = vz[:, h - 1, idx_j]
            vx[:, h - 2, idx_j] = vz[:, h - 2, j_start + 1:j_end + 1] - vz[:, h - 2, idx_j] + vz[:, h - 1, j_start + 1:j_end + 1] - vz[:, h - 1, idx_j] + vx[:, h, idx_j]
            vz[:, h - 3, idx_j] = vz[:, h - 2, idx_j]

        # Apply damping
        vx *= damp
        vz *= damp
        
        # Record receiver waveforms
        rcv_txx[:, t, rcv_idx] = txx[:, rcv_z, rcv_x]
        rcv_tzz[:, t, rcv_idx] = tzz[:, rcv_z, rcv_x]
        rcv_txz[:, t, rcv_idx] = txz[:, rcv_z, rcv_x]
        rcv_vx[:, t, rcv_idx] = vx[:, rcv_z, rcv_x]
        rcv_vz[:, t, rcv_idx] = vz[:, rcv_z, rcv_x]
        
        # Store forward wavefields for visualization or further processing
        forward_wavefield_txx += torch.sum(txx * txx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_tzz += torch.sum(tzz * tzz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_txz += torch.sum(txz * txz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vx += torch.sum(vx * vx, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()
        forward_wavefield_vz += torch.sum(vz * vz, dim=0)[offset:offset + nz, nabc:nabc + nx].detach()

    return txx,tzz,txz,vx,vz,rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz,\
            forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz


##########################################################################
#                       forward Modeling    
##########################################################################
def forward_kernel( nx:int,nz:int,dx:float,dz:float,nt:int,dt:float,
                    nabc:int,free_surface:bool,                                         # Model settings
                    src_x:Tensor,src_z:Tensor,src_n:int,src_v:Tensor,MT:Tensor,         # Source
                    rcv_x:Tensor,rcv_z:Tensor,rcv_n:int,                                # Receiver
                    abc_type:str,bcx:Tensor,bcz:Tensor,damp:Tensor,                     # PML/ABL
                    lamu:Tensor,lam:Tensor,bx:Tensor,bz:Tensor,                         # lame constant
                    CC:List[Tensor],                                                    # elastic moduli
                    fd_order=4,n_segments = 1,                                          # Finite Difference
                    device: torch.device = torch.device("cpu"), dtype=torch.float32
                ):
    # free surface offset
    fs_offset = fd_order//2
    
    # forward simulation
    nx_pml = nx + 2 * nabc
    nz_pml = nz + (nabc + fs_offset if free_surface else 2 * nabc + fs_offset)
    
    # padding input parameter
    lamu = pad_torchSingle(lamu,nabc,fs_offset,free_surface,device)
    lam  = pad_torchSingle(lam ,nabc,fs_offset,free_surface,device)
    bx   = pad_torchSingle(bx  ,nabc,fs_offset,free_surface,device)
    bz   = pad_torchSingle(bz  ,nabc,fs_offset,free_surface,device)
    
    [C11,_,C13,_,C15,_,_,_,_,_,_,C33,_,C35,_,_,_,_,C55,_,_] = CC
    C11 = pad_torchSingle(C11,nabc,fs_offset,free_surface,device)
    C13 = pad_torchSingle(C13,nabc,fs_offset,free_surface,device)
    C15 = pad_torchSingle(C15,nabc,fs_offset,free_surface,device)
    C33 = pad_torchSingle(C33,nabc,fs_offset,free_surface,device)
    C35 = pad_torchSingle(C35,nabc,fs_offset,free_surface,device)
    C55 = pad_torchSingle(C55,nabc,fs_offset,free_surface,device)

    # padding the absorbing boundary condition
    if abc_type.lower() in ["pml"]:
        bcx = pad_torchSingle(bcx,0,fs_offset,free_surface,device)
        bcz = pad_torchSingle(bcz,0,fs_offset,free_surface,device)
    else:
        damp = pad_torchSingle(damp,0,fs_offset,free_surface,device)

    # padding source and receiver
    src_x = src_x + nabc
    src_z = src_z + fs_offset if free_surface else src_z+nabc + fs_offset
    rcv_x = rcv_x + nabc
    rcv_z = rcv_z + fs_offset if free_surface else rcv_z+nabc + fs_offset
    
    # some intermediate variables
    vx,vz,txx,tzz,txz           = torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device)
    vx_x,vz_x,txx_x,tzz_x,txz_x = torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device)
    vx_z,vz_z,txx_z,tzz_z,txz_z = torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device),torch.zeros((src_n,nz_pml,nx_pml),dtype=dtype,device=device)
    rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz = torch.zeros((src_n,nt,rcv_n),dtype=dtype,device=device),torch.zeros((src_n,nt,rcv_n),dtype=dtype,device=device),torch.zeros((src_n,nt,rcv_n),dtype=dtype,device=device),torch.zeros((src_n,nt,rcv_n),dtype=dtype,device=device),torch.zeros((src_n,nt,rcv_n),dtype=dtype,device=device)

    forward_wavefield_txx = torch.zeros((nz,nx),dtype=dtype,device=device)
    forward_wavefield_tzz = torch.zeros((nz,nx),dtype=dtype,device=device)
    forward_wavefield_txz = torch.zeros((nz,nx),dtype=dtype,device=device)
    forward_wavefield_vx  = torch.zeros((nz,nx),dtype=dtype,device=device)
    forward_wavefield_vz  = torch.zeros((nz,nx),dtype=dtype,device=device)
    
    # checkpoints for saving memory
    k = 0
    for i, chunk in enumerate(torch.chunk(src_v,n_segments,dim=-1)):
        if abc_type.lower() in ["pml"]:
            txx_x,txx_z,tzz_x,tzz_z,txz_x,txz_z,txx,tzz,txz,\
            vx_x,vx_z,vz_x,vz_z,vx,vz,\
            rcv_txx_temp,rcv_tzz_temp,rcv_txz_temp,rcv_vx_temp,rcv_vz_temp,\
            forward_wavefield_txx_temp,forward_wavefield_tzz_temp,forward_wavefield_txz_temp,forward_wavefield_vx_temp,forward_wavefield_vz_temp \
                                                                            = checkpoint(step_forward_PML_10order if fd_order == 10 else step_forward_PML_8order if fd_order == 8 else step_forward_PML_6order if fd_order == 6 else step_forward_PML_4order,
                                                                                fd_order,
                                                                                free_surface,nx,nz,dx,dz,nabc,
                                                                                src_x,src_z,src_n,dt,chunk,MT,
                                                                                rcv_x,rcv_z,rcv_n,
                                                                                bcx,bcz,
                                                                                lam,lamu,
                                                                                C11,C13,C15,C33,C35,C55,
                                                                                bx,bz,
                                                                                txx_x,txx_z,tzz_x,tzz_z,txz_x,txz_z,txx,tzz,txz,
                                                                                vx_x,vx_z,vz_x,vz_z,vx,vz,
                                                                                device,dtype,
                                                                                use_reentrant=True)
        else:
            txx,tzz,txz,vx,vz,\
            rcv_txx_temp,rcv_tzz_temp,rcv_txz_temp,rcv_vx_temp,rcv_vz_temp,\
            forward_wavefield_txx_temp,forward_wavefield_tzz_temp,forward_wavefield_txz_temp,forward_wavefield_vx_temp,forward_wavefield_vz_temp \
                                                                            = checkpoint(step_forward_ABL_10order if fd_order == 10 else step_forward_ABL_8order if fd_order == 8 else step_forward_ABL_6order if fd_order == 6 else step_forward_ABL_4order,
                                                                                fd_order,
                                                                                free_surface,nx,nz,dx,dz,nabc,
                                                                                src_x,src_z,src_n,dt,chunk,MT,
                                                                                rcv_x,rcv_z,rcv_n,
                                                                                damp,
                                                                                lam,lamu,
                                                                                C11,C13,C15,C33,C35,C55,
                                                                                bx,bz,
                                                                                txx,tzz,txz,
                                                                                vx,vz,
                                                                                device,dtype,
                                                                                use_reentrant=True)
        # save the waveform recorded on receiver
        rcv_txx[:,k:k+chunk.shape[-1]] = rcv_txx_temp
        rcv_tzz[:,k:k+chunk.shape[-1]] = rcv_tzz_temp
        rcv_txz[:,k:k+chunk.shape[-1]] = rcv_txz_temp
        rcv_vx[:,k:k+chunk.shape[-1]]  =  rcv_vx_temp
        rcv_vz[:,k:k+chunk.shape[-1]]  =  rcv_vz_temp
        
        # save the forward wavefield
        forward_wavefield_txx += forward_wavefield_txx_temp.detach()
        forward_wavefield_tzz += forward_wavefield_tzz_temp.detach()
        forward_wavefield_txz += forward_wavefield_txz_temp.detach()
        forward_wavefield_vx  += forward_wavefield_vx_temp.detach()
        forward_wavefield_vz  += forward_wavefield_vz_temp.detach()
        
        k+=chunk.shape[-1]

    record_waveform = {
        "txx":rcv_txx,
        "tzz":rcv_tzz,
        "txz":rcv_txz,
        "vx" :rcv_vx,
        "vz" :rcv_vz,
        "forward_wavefield_txx":forward_wavefield_txx,
        "forward_wavefield_tzz":forward_wavefield_tzz,
        "forward_wavefield_txz":forward_wavefield_txz,
        "forward_wavefield_vx":forward_wavefield_vx,
        "forward_wavefield_vz":forward_wavefield_vz,
    }
    return record_waveform


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from abc import abstractmethod

class Misfit():
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def forward(self):
        pass


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
import torch

class Misfit_global_correlation(Misfit):
    """ global correlation misfit functions
    
    Paraemters:
    ------------
        obs (Tensors)   : the observed waveform 
        syn (Tensors)   : the synthetic waveform 
    """
    def __init__(self,dt=1) -> None:
        super().__init__()
        self.dt = dt
    
    # One loop for calculate the misfits (traces)
    def forward(self, obs, syn):
        """Compute the global correlation misfit between observed and synthetic waveforms.
        
        Args:
            obs (Tensor): Observed waveform, shape (batch, channels, traces).
            syn (Tensor): Synthetic waveform, shape (batch, channels, traces).
        
        Returns:
            Tensor: Correlation-based misfit loss.
        """
        mask1    = torch.sum(torch.abs(obs),axis=1) == 0
        mask2    = torch.sum(torch.abs(syn),axis=1) == 0
        mask     = ~(mask1 * mask2)
        
        # Initialize result tensor
        rsd = torch.zeros((obs.shape[0], obs.shape[2]), device=obs.device)

        # Compute correlation for each trace
        for itrace in range(obs.shape[2]):
            shot_idx  = torch.argwhere(mask[:,itrace])
            obs_trace = obs[shot_idx, :, itrace].squeeze(axis=1)  # Shape: (N, T)
            syn_trace = syn[shot_idx, :, itrace].squeeze(axis=1)  # Shape: (N, T)

            obs_trace_norm = obs_trace.norm(dim=1, keepdim=True)
            syn_trace_norm = syn_trace.norm(dim=1, keepdim=True)
            
            obs_trace = obs_trace/obs_trace_norm
            syn_trace = syn_trace/syn_trace_norm
            
            # Calculate covariance and variances
            cov     = torch.mean(obs_trace * syn_trace, dim=1)  # Shape: (N,)
            var_obs = torch.var(obs_trace, dim=1)               # Shape: (N,)
            var_syn = torch.var(syn_trace, dim=1)               # Shape: (N,)

            # Avoid division by zero by masking
            corr = cov / (torch.sqrt(var_obs * var_syn) + 1e-8)  # Adding small value to avoid div by zero

            # Handle the case where both variances are zero
            corr[torch.isnan(corr)] = 0  # If both variances are zero, set correlation to zero
            
            rsd[shot_idx, itrace] = -corr.reshape(-1,1)
        
        loss = torch.sum(rsd * self.dt)
        return loss


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-19 22:07:06
* LastEditors: LiuFeng
* LastEditTime: 2024-05-10 14:53:29
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import matplotlib.pyplot as plt
import numpy as np
import os
from mpl_toolkits import axes_grid1

import warnings
warnings.filterwarnings("ignore")

def add_colorbar(im, aspect=20, pad_fraction=0.5, **kwargs):
    """Add a vertical color bar to an image plot."""
    divider = axes_grid1.make_axes_locatable(im.axes)
    width = axes_grid1.axes_size.AxesY(im.axes, aspect=1./aspect)
    pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
    current_ax = plt.gca()
    cax = divider.append_axes("right", size=width, pad=pad)
    plt.sca(current_ax)
    return im.axes.figure.colorbar(im, cax=cax, **kwargs)

# def plot_observeSystem(param,model,src,rcv,save_path="",show=False):
#     v = model.v
#     rcv_x = rcv.rcv_x
#     rcv_z = rcv.rcv_z
#     src_x = src.src_x
#     src_z = src.src_z
#     pml = param.pml
    
#     fig = plt.figure(figsize=(12,8))
#     ax = plt.axes()
#     # im = plt.imshow(v,cmap='jet_r',vmin=3000,vmax=3030)
#     im = plt.imshow(v,cmap='jet_r')
#     cax = fig.add_axes([ax.get_position().x1+0.01,ax.get_position().y0,0.02,ax.get_position().height])
#     # plt.colorbar(im,cax=cax)
#     plt.colorbar(im,cax=cax)
#     ax.scatter(rcv_x,rcv_z,20,marker="v",c='w',label="receiver")
#     ax.scatter(src_x-pml,src_z-pml,20,marker='*',c='k',label="source")
#     ax.legend(fontsize=12)
#     if save_path != "":
#         plt.savefig(save_path,bbox_inches="tight")
#     if show:
#         plt.show()
#     else:
#         plt.close()



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: Thanks to Dr. Deng Bao for modifying this objective function
* Bao Deng (University of Science and Technology of China)
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
import torch
import torch.nn.functional as F

""" 
NOTE!!!

This program implements an experimental traveltime objective function for waveform data
analysis. The following are key considerations and methods used in this approach:

1. **Cross-Correlation Calculation**:
   - The function `cross_correlation` computes the cross-correlation between observed and synthetic waveforms.
   - Cross-correlation is truncated to retain only the center portion of the result to match the length of the original waveform, avoiding the extended size caused by padding.

2. **Non-Differentiability of `argmax`**:
   - Traditional traveltime misfit calculations often rely on finding the peak location of the cross-correlation using `argmax`, which is non-differentiable.
   - This makes it unsuitable for gradient-based optimization methods, as backpropagation requires differentiable operations.

3. **Differentiable Approximation with Softmax**:
   - To make the process compatible with gradient-based methods, we use a softmax-weighted sum over indices in `trvaletime_difference`.
   - Specifically, `tt_loss` approximates the peak location by calculating `sum(softmax * indices)`, where `beta` controls the softness of the weighting. A high beta value sharpens the softmax, mimicking the `argmax` behavior more closely.
   - This alternative, while differentiable, has yet to be fully validated for robustness in noisy or amplitude-varying environments.

4. **Experimental Status**:
   - This objective function remains under development and testing, particularly in challenging scenarios like high noise or amplitude discrepancies.
   - The value of `beta` is adjustable and may need tuning based on specific data properties.

5. **Misfit Calculation**:
   - The `Misfit_traveltime` class normalizes both observed and synthetic waveforms before calculating the residuals.
   - For each source-receiver pair, it computes a traveltime difference and accumulates the absolute values as the total loss.

This code is designed to test the performance of a differentiable approximation for traveltime misfit calculation, though improvements may be needed after further testing and evaluation.
"""

def calculate_time_shift(wave1,wave2,beta=100):
    """Calculate the travel time difference based on cross-correlation."""
    cross_corr = F.conv1d(
        wave1.view(1, 1, -1),
        wave2.view(1, 1, -1),
        padding=wave1.numel()-1
    ).view(-1)
    weights = F.softmax(beta*cross_corr,dim=0)
    tt_lags = torch.arange(-wave1.numel() + 1, wave1.numel(), device=wave1.device, dtype=torch.float32)
    time_shift = torch.sum(weights*tt_lags)
    return time_shift

class Misfit_traveltime(Misfit):
    '''Waveform L2-norm difference misfit (Tarantola, 1984)
    
    Parameters:
    -----------
        dt (float)      : Time sampling interval.
        obs (Tensor)    : Observed waveform.
        syn (Tensor)    : Synthetic waveform.
    '''
    def __init__(self, dt=1, beta=10) -> None:
        super().__init__()
        self.dt = dt
        self.beta = beta
    
    def forward(self, obs: torch.Tensor, syn: torch.Tensor) -> torch.Tensor:
        srcn, nt, rcvn = obs.shape
        padding = nt - 1
        device = obs.device

        # normalization
        obs = obs / torch.max(torch.abs(obs), dim=1, keepdim=True)[0]  # Normalize observed data
        syn = syn / torch.max(torch.abs(syn), dim=1, keepdim=True)[0]  # Normalize synthetic data
        
        rsd = torch.zeros((srcn, rcvn), device=device)  # Reset residual tensor
        for ishot in range(srcn):
            for ircv in range(rcvn):
                # cross_corr = cross_correlation(obs[ishot, :, ircv], syn[ishot, :, ircv], padding=padding)
                # tt_loss = trvaletime_difference(cross_corr=cross_corr, beta=self.beta)*self.dt
                tt_loss = calculate_time_shift(obs[ishot,:,ircv],syn[ishot,:,ircv],beta=self.beta)*self.dt
                rsd[ishot, ircv] = torch.abs(tt_loss)
        
        loss = torch.sum(rsd)  # Compute the total loss
        return loss

# def cross_correlation(wave1, wave2, padding):
#     """Calculate the cross-correlation between two waveforms."""
#     cross_corr = F.conv1d(
#         wave1.view(1, 1, -1),
#         wave2.view(1, 1, -1),
#         padding=padding
#     ).squeeze().abs()
#     center_idx = padding
#     cross_corr = cross_corr[center_idx : center_idx + wave1.size(-1)]
#     return cross_corr

# def trvaletime_difference(cross_corr, beta=100):
#     """Calculate the travel time difference based on cross-correlation."""
#     *_, n = cross_corr.shape
#     input = F.softmax(beta * cross_corr, dim=-1)            # Fix variable reference
#     indices = torch.linspace(0, 1, n, device=input.device)  # Move to the right device
#     tt_loss = torch.sum(n * input * indices, dim=-1)
#     return tt_loss 

# class Misfit_traveltime(torch.autograd.Function):
#     """Implementation of the cross-correlation misfit function
#         s = 0.5*\delta \tau**2
#         where \delta \tau is the time shift between synthetic and observed data
        
#         Luo, Y., & Schuster, G. T. (1991). Wave-equation traveltime inversion.
#             Geophysics, 56(5), 645-653.
#     """
#     @staticmethod
#     def forward(ctx,syn:torch.Tensor,obs:torch.Tensor):
#         """forward pass of the cross-correlation misfit function"""
#         nsrc,nt,nrec    = syn.shape
#         padding         = nt -1 
#         max_time_lags   = torch.zeros((nsrc,nrec),device=syn.device,dtype=syn.dtype)
        
#         # compute the time shift by cross-correlation
#         for isrc in range(nsrc):
#             for irec in range(nrec):
#                 # avoid zero traces
#                 if syn[isrc,:,irec].sum() == 0 or obs[isrc,:,irec].sum() == 0:
#                     continue
#                 cross_corr = F.conv1d(
#                     syn[isrc,:,irec].view(1,1,-1),
#                     obs[isrc,:,irec].view(1,1,-1)
#                 ).squeeze().abs()
#                 max_time_lags[isrc,irec] = torch.argmax(cross_corr) - padding
#         loss = 0.5*torch.sum(max_time_lags**2)
#         # save necessary values for backward pass
#         ctx.save_for_backward(syn,max_time_lags)
#         return loss
    
#     @staticmethod
#     def backward(ctx,grad_output):
#         """backward pass of the cross-correlation travetime misfit function
#             grad_output: the gradient of last layer
            
#             ds/dv = d(delta tau)/dv * delta tau
#                   = 1/(dp_syn/dt) * dp_syn/dv * delta tau
#         """
#         syn,max_time_lags = ctx.saved_tensors
#         # compute the gradient
#         adj = torch.zeros_like(syn)
#         adj[:,1:-1,:] = (syn[:,2:,:] - syn[:,0:-2,:])/2.0
#         adj_sum_squared = torch.sum(adj**2,dim=1,keepdim=True)
        
#         # avoid division by zero
#         adj_sum_squared[adj_sum_squared == 0] = 1.0
        
#         adj = adj/adj_sum_squared
#         adj = adj * (max_time_lags).unsqueeze(1)*grad_output
        
#         # check for NaNs
#         if torch.isnan(adj).any():
#             raise ValueError("NaNs in the gradient")
#         return adj,None


"""
* Modified from: https://github.com/seisfwi/SWIT
* Original Author: Haipeng Li
* Original Author Email: haipeng@stanford.edu
=========================================================
* Author: Liu Feng (SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: Liu Feng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by Liu Feng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
"""

from typing import Optional
from .receiver import Receiver
from .source import Source
from ADFWI.utils import list2numpy
from ADFWI.view import plot_survey
import numpy as np

class Survey(object):
    """Survey class describes the seismic acquisition geometry (2D). I assume 
    that all sources share the same receivers, time samples, and time interval.

    Parameters
    ----------
    source : Source 
        Source object
    receiver : Receiver
        Receiver object
    receiver_mask: The index of useful receiver at each shot
        numpy: [shot number, receiver number], by default None
        this parameter is useful for design the special observation system (line one-side)
                * v v v v v v v v
                    * v v v v v v
                        * v v v v
                            * v v 
                (* means source, v means receiver)
    device : str, optional
        Device for computation: cpu or gpu, by default 'cpu'
    cpu_num : int, optional
        Maximum number of CPU cores, if cpu, by default 1
    gpu_num : int, optional
        Maximum number of GPU cards, if cuda, by default 1
    """
    def __init__(self,source:Source,receiver:Receiver,receiver_masks=None,receiver_masks_obs=True) -> None:
        self.source         = source
        self.receiver       = receiver
        # receive mask  -> mask some of the receiver are useful while other are not
        self.receiver_masks = None
        # receiver_masks_obs -> mark if the obs waveform need to be masked or not
        self.receiver_masks_obs = receiver_masks_obs
        if receiver_masks is not None:
            self.set_receiver_masks(receiver_masks)
    
    def set_receiver_masks(self,receiver_masks):
        src_x,src_z = list2numpy(self.source.loc_x),list2numpy(self.source.loc_z)
        rcv_x,rcv_z = list2numpy(self.receiver.loc_x),list2numpy(self.receiver.loc_z)
        if receiver_masks.shape[0] == len(src_x) and receiver_masks.shape[1] == len(rcv_x):        
            self.receiver_masks = receiver_masks
        else:
            raise ValueError(
                "Receiver Mask Errror: the number of receiver/source are not equal to the Mask"
            )
        
    
    def __repr__(self):
        """ Reimplement the repr function for printing the survey information
        """
        info = f"Survey Information:\n"
        info += repr(self.source)
        info += "\n"
        info += repr(self.receiver)
        return info
    
    def plot(self,model_data,**kwargs):
        src_x = list2numpy(self.source.loc_x)
        src_z = list2numpy(self.source.loc_z)
        rcv_x = list2numpy(self.receiver.loc_x)
        rcv_z = list2numpy(self.receiver.loc_z)
        
        plot_survey(src_x,src_z,rcv_x,rcv_z,model_data,**kwargs)
    
    def plot_single_shot(self,model_data,src_idx,**kwargs):
        src_x = list2numpy(self.source.loc_x[src_idx])
        src_z = list2numpy(self.source.loc_z[src_idx])
        rcv_x = list2numpy(self.receiver.loc_x)
        rcv_z = list2numpy(self.receiver.loc_z)
        if self.receiver_masks is None:
            receiver_mask = np.ones(len(rcv_x))
        else:
            receiver_mask = self.receiver_masks[src_idx]
        rcv_x = rcv_x[np.argwhere(receiver_mask)]
        rcv_z = rcv_z[np.argwhere(receiver_mask)]
        
        plot_survey(src_x,src_z,rcv_x,rcv_z,model_data,**kwargs)
        


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-06-01 14:47:30
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import numpy as np
import torch
from torch import Tensor
from typing import Optional,Tuple,Union
from ADFWI.utils       import gpu2cpu,numpy2tensor
from ADFWI.model.base  import AbstractModel
from ADFWI.view        import (plot_vp_rho,plot_model)
from ADFWI.survey      import Survey

class AcousticModel(AbstractModel):
    """Acoustic Velocity model with parameterization of vp and rho
    """
    def __init__(self,
                ox:float,oz:float,
                nx:int,nz:int,
                dx:float,dz:float,
                vp:Optional[Union[np.array,Tensor]]              = None,     # model parameter
                rho:Optional[Union[np.array,Tensor]]             = None,
                vp_bound: Optional[Tuple[float, float]]          = None,     # model parameter's boundary
                rho_bound: Optional[Tuple[float, float]]         = None,
                vp_grad:Optional[bool]                           = False,    # requires gradient or not
                rho_grad:Optional[bool]                          = False,
                auto_update_rho:Optional[bool]                   = True,
                auto_update_vp :Optional[bool]                   = False,
                water_layer_mask:Optional[Union[np.array,Tensor]]= None,
                free_surface:Optional[bool]                      = False,
                abc_type:Optional[str]                           = 'PML',
                abc_jerjan_alpha:Optional[float]                 = 0.0053,
                nabc:Optional[int]                               = 20,
                device                                           = 'cpu',
                dtype                                            = torch.float32
                )->None:
        """
        Parameters:
        --------------
        ox (float), oz (float)                               : Non use, The origin coordinates of the model in the x- and z- directions (in meters).
        nx (int), nz (int)                                   : Non use, The number of grid points in the x- and z- directions.
        dx (float), dz (float)                               : The grid spacing in the x- and z- directions (in meters).
        vp (Optional[Union[np.array, Tensor]])               : P-wave velocity model with shape (nz, nx). Default is None.
        rho (Optional[Union[np.array, Tensor]])              : Density model with shape (nz, nx). Default is None.
        vp_bound (Optional[Tuple[float, float]])             : The lower and upper bounds for the P-wave velocity model. Default is None.
        rho_bound (Optional[Tuple[float, float]])            : The lower and upper bounds for the density model. Default is None.
        vp_grad (Optional[bool])                             : A flag to indicate if the gradient of the P-wave velocity model is needed. Default is False.
        rho_grad (Optional[bool])                            : A flag to indicate if the gradient of the density model is needed. Default is False.
        auto_update_rho (Optional[bool])                     : Whether to automatically update the density model during inversion. Default is True.
        auto_update_vp (Optional[bool])                      : Whether to automatically update the P-wave velocity model during inversion. Default is False.
        water_layer_mask (Optional[Union[np.array, Tensor]]) : A mask for the water layer (not update), if applicable. Default is None.
        free_surface (Optional[bool])                        : A flag to indicate the presence of a free surface in the model. Default is False.
        abc_type (Optional[str])                             : The type of absorbing boundary condition used in the model. Options include 'PML' and 'Jerjan'. Default is 'PML'.
        abc_jerjan_alpha (Optional[float])                   : The attenuation factor for the Jerjan boundary condition. Default is 0.0053.
        nabc (Optional[int])                                 : The number of grid cells dedicated to the absorbing boundary, default is 20.
        device (str)                                         : The device on which to run the model. Options are 'cpu' or 'cuda'. Default is 'cpu'.
        dtype (torch.dtype)                                  : The data type for PyTorch tensors. Default is torch.float32.
        """
        # initialize the common model parameters
        super().__init__(ox,oz,nx,nz,dx,dz,free_surface,abc_type,abc_jerjan_alpha,nabc,device,dtype)

        # initialize the model parameters
        self.pars       = ["vp","rho"]
        self.vp         = vp.copy()
        self.rho        = rho.copy()
        self.vp_grad    = vp_grad
        self.rho_grad   = rho_grad
        self._parameterization()
        
        # set model bounds
        self.lower_bound["vp"]  =  vp_bound[0]  if vp_bound  is not None else None
        self.lower_bound["rho"] = rho_bound[0]  if rho_bound is not None else None
        self.upper_bound["vp"]  =  vp_bound[1]  if vp_bound  is not None else None
        self.upper_bound["rho"] = rho_bound[1]  if rho_bound is not None else None
        
        # set model gradients
        self.requires_grad["vp"]    = self.vp_grad
        self.requires_grad["rho"]   = self.rho_grad
        
        # check the input model
        self._check_bounds()
        self.check_dims()
        
        # update rho using the empirical function
        self.auto_update_rho = auto_update_rho
        self.auto_update_vp  = auto_update_vp
        
        if water_layer_mask is not None:
            self.water_layer_mask = numpy2tensor(water_layer_mask,dtype=torch.bool).to(device)
        else:
            self.water_layer_mask = None
        
    def _parameterization(self):
        """setting variable and gradients
        """
        # numpy2tensor
        self.vp     = numpy2tensor(self.vp   ,self.dtype).to(self.device)
        self.rho    = numpy2tensor(self.rho  ,self.dtype).to(self.device)
        # set model parameters
        self.vp     = torch.nn.Parameter(self.vp    ,requires_grad=self.vp_grad)
        self.rho    = torch.nn.Parameter(self.rho   ,requires_grad=self.rho_grad)
        return
    
    def get_clone_data(self) -> Tuple:
        """clone the class
        """
        kwargs = super().get_clone_data()
        return kwargs
    
    def _plot_vp_rho(self,**kwargs):
        """plot velocity model
        """
        plot_vp_rho(self.vp,self.rho,
                    dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot(self,var,**kwargs):
        """plot single velocity model
        """
        model_data = self.get_model(var)
        plot_model(model_data,title=var,**kwargs)
        return
    
    def set_rho_using_empirical_function(self):
        """approximate rho via empirical relations with vp
        """
        rho         = self.rho.cpu().detach().numpy()
        vp          = self.vp.cpu().detach().numpy()
        rho_empirical  = np.power(vp, 0.25) * 310
        if self.water_layer_mask is not None:
            mask = self.water_layer_mask.cpu().detach().numpy()
            rho_empirical[mask] = rho[mask]
        rho         = numpy2tensor(rho_empirical,self.dtype).to(self.device)
        self.rho    = torch.nn.Parameter(rho   ,requires_grad=self.rho_grad)
        return

    def set_vp_using_empirical_function(self):
        """approximate vp via empirical relations with rho
        """
        rho         = self.rho.cpu().detach().numpy()
        vp          = self.vp.cpu().detach().numpy()
        vp_empirical= np.power(rho / 310, 4)
        if self.water_layer_mask is not None:
            grad_mask = self.water_layer_mask.cpu().detach().numpy()
            vp_empirical[grad_mask] = vp[grad_mask]
        vp          = numpy2tensor(vp_empirical,self.dtype).to(self.device)
        self.vp     = torch.nn.Parameter(vp , requires_grad=self.vp_grad)
        return   
    
    def clip_params(self)->None:
        """Clip the model parameters to the given bounds
        """
        for par in self.pars:
            if self.lower_bound[par] is not None and self.upper_bound[par] is not None:
                # Retrieve the model parameter
                m = getattr(self, par)
                min_value = self.lower_bound[par]
                max_value = self.upper_bound[par]

                # Create a temporary copy for masking purposes
                m_temp = m.clone()  # Use .clone() instead of .copy() to avoid issues with gradients

                # Clip the values of the parameter using in-place modification with .data
                m.data.clamp_(min_value, max_value)

                # Apply the water layer mask if it is not None, using in-place modification
                if self.water_layer_mask is not None:
                    m.data = torch.where(self.water_layer_mask.contiguous(), m_temp.data, m.data)
        return
        
    def forward(self) -> Tuple:
        """Forward method of the elastic model class
        """
        # using the empirical function to setting rho
        if self.auto_update_rho and not self.rho_grad:
            self.set_rho_using_empirical_function()
        
        if self.auto_update_vp and not self.vp_grad:
            self.set_vp_using_empirical_function()
            
        # Clip the model parameters
        self.clip_params()
        return 


"""
* Author: Liu Feng (SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: Liu Feng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by Liu Feng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
"""

# import numpy as np

# np.random.seed(1234)
# def add_gaussian_noise(data, std_noise, mean_factor=0.1):
#     """
#     Add Gaussian noise with a specified mean (based on each trace) and standard deviation to each trace in the data.

#     Parameters:
#     data (np.ndarray): Input data with dimensions [shot, time, trace].
#     std_noise (float): Standard deviation of the Gaussian noise, controls the noise intensity.
#     mean_factor (float): Factor to scale the mean of the Gaussian noise, controls the noise bias.

#     Returns:
#     np.ndarray: Data with added Gaussian noise.
#     """
#     noise_data = np.zeros_like(data)
#     for ishot in range(data.shape[0]):
#         for itrace in range(data.shape[2]):
#             # Calculate mean_noise based on the mean of the current trace
#             trace_mean = np.mean(data[ishot, :, itrace])
#             mean_noise = trace_mean * mean_factor
            
#             # Add noise independently to each trace with its own mean and standard deviation
#             noise_data[ishot, :, itrace] = data[ishot, :, itrace] + \
#                 np.random.normal(mean_noise, np.std(data[ishot, :, itrace]) * std_noise, data[ishot, :, itrace].shape)
#     return noise_data

import numpy as np

def add_gaussian_noise(data, std_noise, mean_bias_factor=0.1, seed=1234):
    """
    Add Gaussian noise with a specified standard deviation and an optional mean bias to each trace in the data.

    Parameters:
    data (np.ndarray): Input data with dimensions [shot, time, trace].
    std_noise (float): Standard deviation of the Gaussian noise, controls the noise intensity.
    mean_bias_factor (float): Factor to scale the mean of each trace to serve as the mean of the Gaussian noise. 
                              Default is 0.1, meaning the noise mean is 10% of the trace mean.
    seed (int, optional): Random seed for reproducibility. If None, the noise will be different each time.

    Returns:
    np.ndarray: Data with added Gaussian noise.
    """

    # Optionally set the random seed for reproducibility
    if seed is not None:
        np.random.seed(seed)

    # Calculate the mean of each trace [shot, 1, trace] to apply mean bias
    trace_means = np.mean(data, axis=1, keepdims=True)  # Shape: [shot, 1, trace]

    # Calculate the mean noise based on the trace means and the provided mean_bias_factor
    mean_noises = trace_means * mean_bias_factor  # Shape: [shot, 1, trace]
    
    # Generate Gaussian noise with the specified mean bias and standard deviation
    noise = np.random.normal(loc=mean_noises, scale=std_noise, size=data.shape)

    # Add the generated noise to the original data
    noisy_data = data + noise

    return noisy_data



import numpy as np

# source wavelet
def wavelet(nt, dt, f0, amp0 = 1, t0 = None, type = 'Ricker'):
    """ source time function.
    """
    # time array
    t = np.arange(nt) * dt + 0.0
    wavelet = np.zeros_like(t)
    t0 = t0 if t0 is not None else 1.2 / f0

    # Ricker wavelet
    if type.lower() in ['ricker']:
        tau = (np.pi*f0) ** 2
        wavelet = amp0 * (1 - 2 * tau * (t - t0) ** 2) * np.exp(- tau * (t - t0) ** 2)
    
    # Gaussian wavelet
    elif type.lower() in ['gaussian']:
        tau = (np.pi*f0) ** 2
        wavelet = amp0 * (1 - 2 * tau * (t - t0) ** 2) * np.exp(- tau * (t - t0) ** 2)

        # perform integration twice to get the Gaussian wavelet
        wavelet = np.cumsum(wavelet)
        wavelet = np.cumsum(wavelet)

    # Ramp wavelet
    elif type.lower() in ['ramp']:
        wavelet = amp0 * 0.5 * (1. + np.tanh(t / t0))
        
    # Unknown source type
    else:
        msg = 'Support source types: Rikcer, Guassian, Ramp. \n'
        err = 'Unknown source type: {}'.format(type)
        raise ValueError(msg + '\n' + err)

    return t,wavelet